#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import sys
from prettytable import PrettyTable
sys.path.append("../src")

from commands.keepass import KeePass, IStatusLogger, get_homepath, lsdb

class ShowWarningsLogger(IStatusLogger):
    __namespace__ = "KPCLibPy"
    
    def StartLogging(self, strOperation, bWriteOperationToLog):
        print('StartLogging {} {}'.format(strOperation, bWriteOperationToLog))
        return
    def EndLogging(self):
        print('EndLogging')
        return
    def SetProgress(self, uPercent):
        #print('In progress {}'.format(uPercent))
        return True
    def SetText(self, strNewText, lsType):
        print('SetText {} {}'.format(strNewText, lsType))
        return True
    def ContinueWork(self):
        return True

def print_group(group, detail=False):
    if detail:
        s_table = PrettyTable(["Time", "Type", "Name"])
        for gp in group.Groups:
            s_table.add_row([gp.LastModificationTime.ToString(), "Group", gp.get_Name()])
        for entry in group.Entries:
            s_table.add_row([entry.LastModificationTime.ToString(), "Entry", entry.Strings.ReadSafe("Title")])
    else:
        s_table = PrettyTable(["Name"])
        for gp in group.Groups:
            s_table.add_row([gp.get_Name()+"/"])
        for entry in group.Entries:
            s_table.add_row([entry.Strings.ReadSafe("Title")])

    s_table.align = "l"
    s_table.border = False
    s_table.header = False
    print("Current group is {}.\n".format(group))
    print(s_table)

if __name__ == '__main__':
    _keepass = KeePass()
    logger = ShowWarningsLogger()
    db_path = get_homepath() + '/' + "utdb.kdbx"
    _keepass.open(db_path, "12345", logger)
    print_group(_keepass.root_group)



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

from nubia import context
from nubia import exceptions
from nubia import eventbus
from kpclibpy.commands.keepass import KeePass
from kpclibpy import __version__

_keepass = KeePass()

class NubiaPassXYZContext(context.Context):
    def __init__(self, *args, **kwargs):
        self.keepass = _keepass
        self.version = __version__
        self._current_group = None
        super().__init__()

    @property
    def db_type(self):
        """
        Getter of database type
        The database type can be KeePass or PassXYZ
        """
        return self.keepass.db_type

    @property
    def current_group(self):
        """
        Getter of the current working directory (group)
        """
        if self.keepass.current_group:
            self._current_group = self.keepass.current_group.Name
        else:
            self._current_group = "Offline"
            
        return self._current_group

    def on_connected(self, *args, **kwargs):
        pass

    def on_cli(self, cmd, args):
        # dispatch the on connected message
        self.verbose = args.verbose
        self.registry.dispatch_message(eventbus.Message.CONNECTED)

    def on_interactive(self, args):
        self.verbose = args.verbose
        ret = self._registry.find_command("connect").run_cli(args)
        if ret:
            raise exceptions.CommandError("Failed starting interactive mode")
        # dispatch the on connected message
        self.registry.dispatch_message(eventbus.Message.CONNECTED)



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

from . import kpclib



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import sys
import os
import pytest

import commands
from nubia import Nubia, Options
from nubia_plugin import NubiaPassXYZPlugin
from commands.keepass import get_homepath

class PxShell(Nubia):
    """
    Unit test class
    >>> import tests.db_test as db_test
    >>> db_test.test_version()

    or
    >>> shell = db_test.PxShell()
    >>> shell.run_interactive_line('version')
    """
    def __init__(self, name="passxyz_test"):
        px_plugin = NubiaPassXYZPlugin()
        super(PxShell, self).__init__(name, plugin=px_plugin, command_pkgs=commands, testing=True)
        self.registry = self._registry
        self.ut_db = "utdb.kdbx"
        self.ut_password = "12345"


    def run_cli_line(self, raw_line):
        cli_args_list = raw_line.split()
        args = self._pre_run(cli_args_list)
        return self.run_cli(args)

    def run_interactive_line(self, raw_line, cli_args=None):
        cli_args = cli_args or "test_shell connect"
        cli_args_list = cli_args.split()
        args = self._pre_run(cli_args_list)
        io_loop = self._create_interactive_io_loop(args)
        return io_loop.parse_and_evaluate(raw_line)

shell = PxShell()

@pytest.fixture(scope='session')
def pxshell():
    global shell
    return shell

@pytest.fixture()
def keepass_db(pxshell):
    pxshell.run_interactive_line('open dbfile={} password={}'.format(shell.ut_db, shell.ut_password))
    yield
    pxshell.run_interactive_line('close')



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import clr
import sys

from kpclibpy import __path__ as kpclibpy_path

def interop_dll_path(dll_name):
    # Unfrozen path
    dll_path = os.path.join(os.path.dirname(os.path.realpath(kpclibpy_path[0])), 'kpclibpy', 'kpclib', 'lib', dll_name)
    #print("1. ", dll_path)
    if os.path.exists(dll_path):
        return dll_path

    # Frozen path, dll in the same dir as the executable
    dll_path = os.path.join(os.path.dirname(os.path.realpath(kpclibpy_path[0])), 'kpclibpy', '..', 'kpclib', 'lib', dll_name)
    #print("2. ", dll_path)
    if os.path.exists(dll_path):
        return dll_path

    raise Exception('Cannot find %s' % dll_path)

clr.AddReference(interop_dll_path('KPCLib.dll'))
clr.AddReference(interop_dll_path('SkiaSharp.dll'))
clr.AddReference(interop_dll_path('PassXYZLib.dll'))

if __name__ == '__main__':
	from KeePassLib import PwDatabase, PwGroup
	print(PwDatabase)
	print(PwGroup)



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

from pygments.token import Token

from nubia import context
from nubia import statusbar


class NubiaPassXYZStatusBar(statusbar.StatusBar):
    def __init__(self, context):
        self._last_status = None

    def get_rprompt_tokens(self):
        if self._last_status:
            return [(Token.RPrompt, "Error: {}".format(self._last_status))]
        return []

    def set_last_command_status(self, status):
        self._last_status = status

    def get_tokens(self):
        spacer = (Token.Spacer, "  ")
        if context.get_context().keepass.is_hidden:
            is_verbose = (Token.Warn, "Hidden")
        else:
            is_verbose = (Token.Info, "Show")

        return [
            (Token.Info, context.get_context().db_type),
            spacer,
            (Token.Toolbar, "Password"),
            spacer,
            is_verbose,
            spacer,
            (Token.Toolbar, "Status"),
            spacer,
            (Token.Info, context.get_context().current_group),
        ]



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#
__all__ = (
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
)

__copyright__ = "Copyright 2021 Roger Ye"

import importlib_metadata

try:
    metadata = importlib_metadata.metadata("kpclibpy")


    __title__ = metadata["name"]
    __summary__ = metadata["summary"]
    __uri__ = metadata["home-page"]
    __version__ = metadata["version"]
    __author__ = metadata["author"]
    __email__ = metadata["author-email"]
    __license__ = metadata["license"]
except ModuleNotFoundError:
    __version__ = "1.1.5.dev0"
    print("Debug build ", __version__)


#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

from pathlib import *
from os import listdir, path
from os.path import isfile, join
from prettytable import PrettyTable
from termcolor import cprint, colored
import configparser
import consolemd

import kpclibpy.kpclib

from KeePassLib import PwGroup, PwEntry, SearchParameters
from KeePassLib.Collections import PwObjectList
from KeePassLib.Interfaces import IStatusLogger
from KeePassLib.Keys import CompositeKey, KcpPassword, InvalidCompositeKeyException
from KeePassLib.Security import ProtectedString
from KeePassLib.Serialization import IOConnectionInfo
from PassXYZLib import PxDefs, PxDatabase, PxLibInfo


def get_homepath():
    config_file = str(Path.home())+'/.kpclibdb/kpclibpy.ini'

    if path.exists(config_file):
        config = configparser.ConfigParser()
        config.read(config_file)
        return config['DEFAULT']['homepath']
    else:
        return str(Path.home())+'/.kpclibdb'


def lsdb():
    import fnmatch
    home_path = get_homepath()
    files = fnmatch.filter(listdir(home_path), "*.kdbx") + fnmatch.filter(listdir(home_path), "*.xyz")
    #onlyfiles = [f for f in listdir(home_path) if isfile(join(home_path, f))]
    return files


class KeePass:
    """
    Python interface of KeePassLib
    """
    TITLE = "Title"
    USERNAME = "UserName"
    PASSWORD = "Password"
    URL = "URL"
    NOTES = "Notes"

    _keepass = None

    def __new__(cls, *args, **kwargs):
        if not cls._keepass:
            cls._keepass = super(KeePass, cls
                ).__new__(cls, *args, **kwargs)
        return cls._keepass
    
    def __init__(self):
        self._db = None
        self._current_group = None
        self._db_type = "KeePass"
        self._file_name = None
        self._user_name = None
        self.is_hidden = True

    @property
    def version(self):
        """
        ReadOnly property to represent the version of PassXYZLib
        """
        return PxLibInfo.Version

    @property
    def db(self):
        """
        ReadOnly property to represent a KeePass database instance (PxDatabase)
        """
        return self._db

    @property
    def user_name(self):
        """
        Getter of decoded file name
        """
        if self._file_name:
            self._user_name = PxDefs.GetUserNameFromDataFile(self._file_name)

        return self._user_name

    @property
    def file_name(self):
        """
        Getter of database file name
        """
        return self._file_name

    @file_name.setter
    def file_name(self, name):
        """
        Setter of database file name
        """
        self._file_name = name

    @property
    def db_type(self):
        """
        Getter of database type
        The database type can be KeePass or PassXYZ
        """
        if self._file_name:
            self._db_type = PxDefs.GetDatabaseType(self._file_name)
        return self._db_type

    @db_type.setter
    def db_type(self, db_t):
        """
        Setter of database type
        """
        if db_t == "KeePass" or db_t == "PassXYZ":
            self._db_type = db_t
        else:
            print("Unsupported database type")

    @property
    def  root_group(self):
        """
        Root group of the current database
        """
        return self._db.RootGroup

    @property
    def current_group(self):
        """
        Getter of the current working directory (group)
        """
        if self._db:
            return self._db.CurrentGroup
        else:
            return self._current_group

    @current_group.setter
    def current_group(self, group):
        """
        Setter of the current working directory (group)
        """
        if self._db:
            self._db.CurrentGroup = group
        else:
            self._current_group = group

    @property
    def groups(self):
        """
        A dictionary of groups in current working directory (group)
        Key   - group name
        Value - PwGroup
        """
        if self.current_group:
            groups = {}
            for gp in self.current_group.Groups:
                groups[gp.get_Name()] = gp
            return groups
        else:
            return None

    @property
    def entries(self):
        """
        A dictionary of entries in current working directory (group)
        Key   - entry title
        Value -  PwEntry
        """
        if self.current_group:
            entries = {}
            for entry in self.current_group.Entries:
                entries[entry.Strings.ReadSafe(KeePass.TITLE)] = entry
            return entries
        else:
            return None

    def find(self, keywords):
        sp = SearchParameters()
        sp.SearchString = keywords
        ol = PwObjectList[PwEntry]()
        self.root_group.SearchEntries(sp, ol)
        entries = {}
        for entry in ol:
            entries[entry.Name] = entry
        return entries

    def get_entries(self, name):
        """
        Get a dictionary of entries in the sub-group of current group
        name   - sub-group name
        """
        try:
            if self.groups[name]:
                entries = {}
                for entry in self.groups[name].Entries:
                    entries[entry.Strings.ReadSafe(KeePass.TITLE)] = entry
                return entries
            else:
                return None
        except KeyError:
            return None

    def _add_entry(self, title = "New Entry", username = "", pw = "", url = "", notes = ""):
        entry = PwEntry()
        entry.Name = title
        if username:
            entry.Strings.Set(self.USERNAME, ProtectedString(False, username))
        else:
            entry.Strings.Set(self.USERNAME, ProtectedString(False, self.db.DefaultUserName))
        
        if pw:
            entry.Strings.Set(self.PASSWORD, ProtectedString(True, pw))
        else:
            entry.Strings.Set(self.PASSWORD, ProtectedString(True, self.db.DefaultUserName))

        if url:
            entry.Strings.Set(self.URL, ProtectedString(False, url))

        if notes:
            entry.Strings.Set(self.NOTES, ProtectedString(False, notes))

        self.current_group.AddEntry(entry, True)

    def add_entry(self, title = "New Entry", username = "", pw = "", url = "", notes = ""):
        """
        Added a new entry in the current group
        name   - sub-group name
        """
        try:
            if(self.entries[title]):
                cprint("Entry exist.", "red")
            else:
                self._add_entry(title, username, pw, url, notes)
        except KeyError:
            self._add_entry(title, username, pw, url, notes)

    def update_entry(self, entry, key, value, protected=False):
        """
        Update an entry
        entry  - Entry to be updated
        key   - a key name
        value - a value
        """
        # If it is a PxEntry, we need to encode key
        if PxDefs.IsPxEntry(entry):
            PxDefs.UpdatePxEntry(entry, key, value, protected)
        else:
            entry.Strings.Set(key, ProtectedString(protected, value))


    def get_groups(self, name):
        """
        Get a dictionary of groups in the sub-group of current group
        name   - sub-group name
        """
        try:
            if self.groups[name]:
                groups = {}
                for gp in self.groups[name].Groups:
                    groups[gp.get_Name()] = gp
                return groups
            else:
                return None
        except KeyError:
            return None
        
    def find_group_by_path(self, path):
        return self.db.FindByPath[PwGroup](path)

    def find_entry_by_path(self, path):
        return self.db.FindByPath[PwEntry](path)

    def get_value(self, entry, key):
        if(entry.Strings.GetSafe(key).IsProtected):
            if self.is_hidden:
                return "*******"
        return entry.Strings.ReadSafe(key)

    def print_entry(self,  entry):
        en_table = PrettyTable([colored(KeePass.TITLE, "yellow"), entry.Strings.ReadSafe(KeePass.TITLE)])
        en_table.align = "l"
        en_table.border = False

        for x in entry.Strings:
            if x.Key != KeePass.TITLE and x.Key != KeePass.NOTES:
                value = self.get_value(entry, x.Key)
                if value:
                    if PxDefs.IsPxEntry(entry):
                        en_table.add_row([colored(PxDefs.DecodeKey(x.Key), "yellow"), self.get_value(entry, x.Key)])
                    else:
                        en_table.add_row([colored(x.Key, "yellow"), self.get_value(entry, x.Key)])
        print(en_table)
        #print(entry.Strings.ReadSafe(KeePass.NOTES))
        text = entry.Strings.ReadSafe(KeePass.NOTES)
        kwargs = {}
        renderer = consolemd.Renderer()
        renderer.render(text, **kwargs)

    def is_open(self):
        if self.db:
            if self.db.IsOpen:
                return True
            else:
                return False
        else:
            return False

    def close(self):
        if self._db:
            if self._db.IsOpen:
                self._db.Close()
                self._db = None
        else:
            print("Database connection is not established yet.")

    def connect(self, filename, password):
        """
        This command is used to connect a database.
        connect supports device lock.
        """
        self._db = PxDatabase()
        try:
            self._db.DefaultFolder = get_homepath()
            self._db.Open(filename, password)
        except InvalidCompositeKeyException:
            self.close()
            self._db = None
            cprint("open: The composite key is invalid!", "red")
            return

    def open(self, db_path, password, logger):
        ioc = IOConnectionInfo.FromPath(db_path)
        cmpKey = CompositeKey()
        cmpKey.AddUserKey(KcpPassword(password))

        self._db = PxDatabase()

        try:
            if not self._db.IsOpen:
                self._db.Open(ioc, cmpKey, logger)
                self._file_name = db_path
                #self.current_group = self._db.RootGroup
        except InvalidCompositeKeyException:
            self.close()
            self._db = None
            cprint("open: The composite key is invalid!", "red")
            return

    def new(self, db_path, password):
        ioc = IOConnectionInfo.FromPath(db_path)
        cmpKey = CompositeKey()
        cmpKey.AddUserKey(KcpPassword(password))

        self._db = PxDatabase()

        if not self._db.IsOpen:
            self._db.New(ioc, cmpKey)
            self.current_group = self._db.RootGroup
        else:
            cprint("new: Database already exist.", "red")

    def move_group(self, src_group, dst_group):
        if self._db.IsOpen:
            if self._db.MoveGroup(src_group, dst_group):
                print("Moved {} to {}".format(src_group.get_Name(), dst_group.get_Name()))
            else:
                print("Cannot move {} to {}".format(src_group.get_Name(), dst_group.get_Name()))

    def move_entry(self, src_entry, dst_group):
        if self._db.IsOpen:
            if self._db.MoveEntry(src_entry, dst_group):
                print("Moved {} to {}".format(src_entry.get_Name(), dst_group.get_Name()))
            else:
                print("Cannot move {} to {}".format(src_entry.get_Name(), dst_group.get_Name()))


#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import sys
import os
from typing import Any

import kpclibpy.commands
from nubia import Nubia, Options
from kpclibpy.nubia_plugin import NubiaPassXYZPlugin
from kpclibpy.commands.keepass import get_homepath

def main() -> Any:
    homepath = get_homepath()
    if not os.path.exists(homepath):
        os.mkdir(homepath)

    plugin = NubiaPassXYZPlugin()
    shell = Nubia(
        name="nubia_passxyz",
        command_pkgs=kpclibpy.commands,
        plugin=plugin,
        options=Options(
            persistent_history=False, auto_execute_single_suggestions=False
        ),
    )
    sys.exit(shell.run())

if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import typing
import click
#from pathlib import *
from termcolor import cprint
from nubia import command, argument, context
#from commands.keepass import KeePass, IStatusLogger, get_homepath, lsdb


@command
@argument("path", description="the entry path", positional=True)
def cat(path: str):
    "Show an entry"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        entry = ctx.keepass.find_entry_by_path(path)
        if entry:
            ctx.keepass.print_entry(entry)
        else:
            print("cannot access {}: No such file or directory".format(path))


@command
@argument("path", description="entry path", positional=True)
@argument("key", description="source entry/group")
@argument("value", description="destination entry/group")
@argument("protected", description="print detail information or not")
def edit(path, key="Notes", value="", protected=False):
    "Edit an entry. Need to provide a key and a value to edit a field."
    ctx = context.get_context()
    if ctx.keepass.is_open():
        entry = ctx.keepass.find_entry_by_path(path)
        if entry:
            if value:
                ctx.keepass.update_entry(entry, key, value, protected)
            else:
                value = click.edit(entry.Strings.ReadSafe(key))
                ctx.keepass.update_entry(entry, key, value, protected)
        else:
            print("cannot access {}: No such file or directory".format(path))
    

@command
@argument("title", description="enter a title")
@argument("username", description="enter a username")
@argument("password", description="enter a password")
@argument("url", description="enter a url")
@argument("notes", description="enter a note")
def new(title="New entry", username="", password="", url="", notes=""):
    "Create a new entry"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        ctx.keepass.add_entry(title, username, password, url, notes)


@command
@argument("entry_name", description="enter a entry name", positional=True)
@argument("key", description="source entry/group")
def rm(entry_name: str, key=""):
    "Remove an entry or a field"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        try:
            entry = ctx.keepass.entries[entry_name]
            if entry:
                if key:
                    ctx.keepass.update_entry(entry, key, "")
                    cprint("Removed field {}.".format(key))
                else:
                    ctx.keepass.db.DeleteEntry(entry)
                    cprint("Removed entry {}.".format(entry_name))
            else:
                cprint("rm: cannot remove {}: No such entry".format(entry_name), "red")
        except KeyError:
            cprint("rm: cannot remove {}: No such entry".format(entry_name), "red")



#!/usr/bin/env python

from setuptools import setup, Command, Extension, find_packages
from setuptools.command.build_ext import build_ext

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="kpclibpy",
    #version="1.1.0",
    description="KeePass command line application",
    url="https://github.com/passxyz/KPCLibPy",
    license="MIT",
    author="Roger Ye",
    author_email="shugaoye@yahoo.com",
    install_requires=[
        "pycparser==2.20",
        "consolemd >= 0.5.1",
        "importlib_metadata >= 3.6",
        "pythonnet==2.5.2",
        "python-nubia==0.2b5"],
    long_description=long_description,
    use_scm_version=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: C#",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
    ],
)



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import typing
from prettytable import PrettyTable
from termcolor import cprint, colored
from nubia import command, argument, context
from kpclibpy.commands.keepass import IStatusLogger, get_homepath, lsdb
#from KeePassLib import PwGroup, PwEntry, Collections


class KPCLibPyLogger(IStatusLogger):
    __namespace__ = "KPCLibPyLogger"
    
    def StartLogging(self, strOperation, bWriteOperationToLog):
        print('StartLogging {} {}'.format(strOperation, bWriteOperationToLog))
        return
    def EndLogging(self):
        print('EndLogging')
        return
    def SetProgress(self, uPercent):
        #print('In progress {}'.format(uPercent))
        return True
    def SetText(self, strNewText, lsType):
        print('SetText {} {}'.format(strNewText, lsType))
        return True
    def ContinueWork(self):
        return True

"""
@command("export")
def export_db():
    "Export entries to a new database"
    return None

@command("import")
def import_db():
    "Import another database to the current one"
    return None
"""


def print_entries(entries):
    e_table = PrettyTable(["No", "Entry Name"])
    num = 0
    indexes = {}
    for entry in entries:
        num = num + 1
        e_table.add_row([num, entry])
        indexes[num] = entry
    e_table.align = "l"
    e_table.border = False
    e_table.header = False
    print(e_table)
    return indexes

@command
@argument("keywords", description="Search keyword in entries", positional=True)
def find(keywords: str):
    "Find entries"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        entries = ctx.keepass.find(keywords)
        while True:
            indexes = print_entries(entries)
            x = input('Enter the entry number (or q to quite):')
            try:
                if x == 'q':
                    break
                else:
                    ctx.keepass.print_entry(entries[indexes[int(x)]])
                    input('Please any key to continue')
            except ValueError:
                print("Wrong entry number, please choose a correct one or enter 'q' to quit")
    else:
        cprint("Please connect to a database first.", "red")


def _move_group(src, dst_group):
    """
    Move src to a group
    src       - source path of an entry or a group
    dst_group - a PwGroup instance
    """
    ctx = context.get_context()
    if ctx.keepass.is_open():
        if src:
            src_group = ctx.keepass.find_group_by_path(src)
            if src_group:
                ctx.keepass.move_group(src_group, dst_group)
            else:
                src_entry = ctx.keepass.find_entry_by_path(src)
                if src_entry:
                    ctx.keepass.move_entry(src_entry, dst_group)
                else:
                    print("mv: cannot access {}: No such file or directory".format(src))
        else:
            print("mv: src is empty.")

@command
@argument("src", description="source entry/group", positional=True)
@argument("dst", description="destination group", positional=True)
def mv(src: str, dst: str):
    "Move an entry or group to a new location"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        if dst:
            dst_group = ctx.keepass.find_group_by_path(dst)
            if dst_group:
                _move_group(src, dst_group)
            else:
                print("mv: cannot access group {}: No such file or directory".format(dst))
        else:
            print("mv: dst is empty.")


def rename_entry(src, dst):
    ctx = context.get_context()
    try:
        if(ctx.keepass.entries[src]):
            ctx.keepass.entries[src].set_Name(dst)
            return ctx.keepass.entries[dst]
        else:
            return None
    except KeyError:
        return None

def rename_group(src, dst):
    ctx = context.get_context()
    try:
        if(ctx.keepass.groups[src]):
            ctx.keepass.groups[src].set_Name(dst)
            return ctx.keepass.groups[dst]
        else:
            return None
    except KeyError:
        return None

@command
@argument("src", description="source entry/group", positional=True)
@argument("dst", description="destination entry/group", positional=True)
def rename(src: str, dst: str):
    "Rename a group or entry"
    if rename_entry(src, dst):
        cprint("Renamed entry {} to {}".format(src, dst))
    elif rename_group(src, dst):
        cprint("Renamed group {} to {}".format(src, dst))
    else:
        cprint("rename: cannot stat {}: No such entry or group".format(src))


@command
def version():
    "Display version number"
    en_table = PrettyTable(["Components", "Version"])
    en_table.align = "l"

    ctx = context.get_context()
    en_table.add_row([colored("KPCLibPy", "yellow"), ctx.version])
    en_table.add_row([colored("KPCLib", "yellow"), "{}".format(ctx.keepass.version)])

    if ctx.keepass.is_open():
        en_table.add_row([colored("CurrentGroup", "yellow"), "{}".format(ctx.keepass.current_group)])
    
    print(en_table)
    return ctx.version


@command
def save():
    """
    Save te database to disk
    """
    ctx = context.get_context()
    db = ctx.keepass.db

    if ctx.keepass.is_open():
        logger = KPCLibPyLogger()
        db.Save(logger)


@command("show")
@argument("items", type=str, description="Database configuration, version etc.",
    choices=["config", "version", "databases", "help"], positional=True)
def show(items):
    """
    This command shows the information of the current databases.
    """
    ctx = context.get_context()
    db = ctx.keepass.db

    help_table = PrettyTable(["Sub-command", "Description"])
    help_table.align = "l"
    help_table.add_row(["databases", "list of databases"])
    help_table.add_row(["config", "database configuration"])
    help_table.add_row(["version", "App version"])

    if items == "config":
        if ctx.keepass.is_open():
            db_table = PrettyTable(["Property", "Value"])
            db_table.align = "l"

            db_table.add_row([colored("Database name", "magenta"), db.Name])
            db_table.add_row([colored("DefaultUserName", "magenta"), db.DefaultUserName])
            db_table.add_row([colored("Modified", "magenta"), db.Modified])
            db_table.add_row([colored("Hide password", "magenta"), ctx.keepass.is_hidden])
            db_table.add_row([colored("RecycleBinEnabled", "magenta"), db.RecycleBinEnabled])
            db_table.add_row([colored("HistoryMaxItems", "magenta"), db.HistoryMaxItems])
            db_table.add_row([colored("HistoryMaxSize", "magenta"), db.HistoryMaxSize])
            db_table.add_row([colored("Description", "magenta"), db.Description])
            db_table.add_row([colored("MaintenanceHistoryDays", "magenta"), db.MaintenanceHistoryDays])
            db_table.add_row([colored("Filename", "magenta"), ctx.keepass.file_name])
            if ctx.keepass.db_type == "PassXYZ":
                db_table.add_row([colored("Decoded Filename", "magenta"), ctx.keepass.user_name])
            cprint("Database configureation", "yellow")
            print(db_table)
        else:
            cprint("Database is closed", "red")
    elif items == "version":
        version()
    elif items == "databases":
        print("The current path: ", get_homepath())
        db_table = PrettyTable(["No.", "Name"])
        db_table.align = "l"
        num = 1
        for file in lsdb():
            db_table.add_row([num, file])
            num = num + 1
        print(db_table)
    elif items == "help":
        print(help_table)
    else:
        print(help_table)


@command
@argument("dbfile", description="Pick a data file", choices=lsdb())
@argument("password", description="Please provide your password")
def open(dbfile: str, password = ""):
    """
    This command is used to connect a database.
    """
    if not password:
        import getpass
        password = getpass.getpass('Password:')

    ctx = context.get_context()
    db = ctx.keepass.db

    if ctx.keepass.is_open():
        cprint("Database name: {}\nDescription: {}\nMaintenanceHistoryDays:{}"
            .format(db.Name, db.Description, db.MaintenanceHistoryDays))
    else:
        #logger = KPCLibPyLogger()
        #db_path = get_homepath() + '/' + dbfile
        #ctx.keepass.open(db_path, password, logger)
        ctx.keepass.connect(dbfile, password)
    
    ctx.keepass.file_name = dbfile


@command
@argument("dbfile", description="Pick a data file")
@argument("password", description="Please provide your password")
@argument("type", description="kdbx or xyz")
def newdb(dbfile: str, password = "", type = ".kdbx"):
    """
    This command is used to create a new database.
    """

    ctx = context.get_context()
    db = ctx.keepass.db

    if ctx.keepass.is_open():
        cprint("Please close the current database {} before you can create a new database."
            .format(db.Name))
    else:
        newfile = dbfile + type
        for file in lsdb():
            if newfile == file:
                print("Data file {} already exist.".format(newfile))
                return
        if not password:
            import getpass
            password = getpass.getpass('Password:')
        db_path = get_homepath() + '/' + newfile
        ctx.keepass.new(db_path, password)
        ctx.keepass.db.Name = dbfile
        ctx.keepass.db.DefaultUserName = dbfile
        ctx.keepass.db.Description = db_path


@command
def close():
    """
    Close the database
    """
    ctx = context.get_context()
    ctx.keepass.close()
    cprint("Database is closed")

@command
def clear():
    '''
    Clears console.
    '''
    print("\033c")

# instead of replacing _ we rely on camelcase to - super-command


@command
class Configure:
    "KeePass Database configuration"

    def __init__(self) -> None:
        ctx = context.get_context()
        self.keepass = ctx.keepass
        self._pwdb = ctx.keepass.db
        if not self._pwdb:
            cprint("Please connect to a database first.", "red")

    @property
    def pwdb(self):
        return self._pwdb

    """This is the super command help"""

    @command("name")
    @argument("db_name", positional=False)
    def db_name(self, db_name=""):
        """
        show or change the database name
        """
        if self.pwdb:
            if db_name:
                self.pwdb.Name = db_name
                self.pwdb.Modified = True
                cprint("Updated database name to {}".format(self.pwdb.Name))
            else:
                cprint("Database name: {}".format(self.pwdb.Name))

    @command
    @argument("username", positional=False)
    def default_user_name(self, username=""):
        """
        show or change the default username
        """
        if self.pwdb:
            if username:
                self.pwdb.DefaultUserName = username
                self.pwdb.Modified = True
                cprint("Updated default username to {}".format(self.pwdb.DefaultUserName))
            else:
                cprint("Database name: {}".format(self.pwdb.DefaultUserName))

    @command
    def toggle(self):
        """
        Toggle show or hide password
        """
        if self.keepass:
            if self.keepass.is_hidden:
                self.keepass.is_hidden = False
            else:
                self.keepass.is_hidden = True



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import typing
from prettytable import PrettyTable
from termcolor import cprint
from nubia import command, argument, context
from kpclibpy.commands.keepass import KeePass

def print_group(group, detail=False):
    ctx = context.get_context()

    if detail:
        s_table = PrettyTable(["Time", "Type", "Name"])
        for gp in group.Groups:
            s_table.add_row([gp.LastModificationTime.ToString(), "Group", gp.get_Name()])
        for entry in group.Entries:
            s_table.add_row([entry.LastModificationTime.ToString(), "Entry", entry.Strings.ReadSafe(KeePass.TITLE)])
    else:
        s_table = PrettyTable(["Name"])
        for gp in group.Groups:
            s_table.add_row([gp.get_Name()+"/"])
        for entry in group.Entries:
            s_table.add_row([entry.Strings.ReadSafe(KeePass.TITLE)])

    s_table.align = "l"
    s_table.border = False
    s_table.header = False
    print("Current group is {}.\n".format(group))
    print(s_table)

@command
@argument("path", description="a specified path", positional=False)
@argument("detail", description="print detail information or not")
def ls(path="", detail=False):
    "Lists entries or groups in pwd or in a specified path"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        if path:
            current_group = ctx.keepass.find_group_by_path(path)
            if current_group:
                print_group(current_group, detail)
            else:
                entry = ctx.keepass.find_entry_by_path(path)
                if entry:
                    print("{}".format(entry.get_Name()))
                else:
                    print("cannot access {}: No such file or directory".format(path))
        else:
            print_group(ctx.keepass.current_group, detail)


@command
def pwd():
    "Print the current working directory"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        print(ctx.keepass.db.CurrentPath)


@command
@argument("path", description="enter a path", positional=True)
def cd(path: str):
    """
    Change directory (group), '/' - root, '..' - parent
    """
    ctx = context.get_context()
    if ctx.keepass.is_open():
        if path == '/':
            ctx.keepass.current_group = ctx.keepass.root_group
        else:
            group = ctx.keepass.find_group_by_path(path)
            if group:
                ctx.keepass.current_group = group
            else:
                cprint("Cannnot find {}".format(path), "yellow")


@command
@argument("group_name", description="enter a group name", positional=True)
def mkdir(group_name: str):
    "Create a new directory (group)"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        new_group = ctx.keepass.current_group.FindCreateGroup(group_name, False)
        if new_group:
            cprint("{} already exist.".format(new_group))
        else:
            new_group = ctx.keepass.current_group.FindCreateGroup(group_name, True)
            cprint("Created group {}".format(new_group))

@command
@argument("group_name", description="enter a group name", positional=True)
def rmdir(group_name: str):
    "Delete a directory (group)"
    ctx = context.get_context()
    if ctx.keepass.is_open():
        group = ctx.keepass.current_group.FindCreateGroup(group_name, False)
        if group:
            ctx.keepass.db.DeleteGroup(group)
            cprint("Removed {}.".format(group))
        else:
            cprint("rmdir: failed to remove {}: No such group.".format(group))



#!/usr/bin/env python3

# Copyright (c) Roger Ye.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
#

import pytest

@pytest.mark.order(1)
def test_version(pxshell):
    version = pxshell.run_interactive_line('version')
    assert version == "1.1.0"


@pytest.mark.order(2)
def test_cat(keepass_db, pxshell):
    pxshell.run_interactive_line('cat "Sample Entry"')


@pytest.mark.order(3)
def test_show1(pxshell):
    pxshell.run_interactive_line('show')


@pytest.mark.order(4)
def test_show2(keepass_db, pxshell):
    pxshell.run_interactive_line('show')

